/*
 * tconsole
 * Copyright 1990 Tom Lawrence, Brown University
 * Last Modification 6/14/90
 */

#include <sys/types.h>

#include <sundev/kbd.h>
#include <sundev/kbio.h>

#include <pixrect/pixrect_hs.h>

#include <ctype.h>
#include <stdio.h>

#include "get_char.h"
#include "global.h"
#include "ascii.h"

int STANDOUT_MODE;
int SCROLL_MODE;

Pixrect *sp;
Pixfont *font;

/**********************************************************************
**********************************************************************/

char font_dir[] = "/usr/lib/fonts/fixedwidthfonts";
char fb[] = "/dev/fb";

create_win()
{
	register int i;
	int width, height;
	char *font_real;
	struct winsize winsize;

	if ((sp = pr_open(fb)) == 0) {
		fputs("tconsole: open: ", stderr);
		perror(fb);
		exit(1);
	}
	width = sp->pr_size.x;
	height = sp->pr_size.y;

	if (*FONT == '/')	/* absolute pathname */
		font_real = FONT;
	else {
		i = sizeof(font_dir) - 1;
		++i;
		i += strlen(FONT);
		++i;
		font_real = (char *) malloc(i);
		(void)sprintf(font_real, "%s/%s", font_dir, FONT);
	}

	if ((font = pf_open(font_real)) == 0) {
		fputs("tconsole: pf_open: ", stderr);
		perror(font_real);
		exit(1);
	}
	font_width = font->pf_defaultsize.x;
	font_height = font->pf_defaultsize.y;

	if (default_window) {
		/* Defaults for row/cols */
		if (width == 1600 && height == 1280) {
			/* Hi res */
			if (ioctl(0, TIOCGWINSZ, &winsize) < 0) {
				perror("tconsole: TIOCGWINSZ");
				exit(1);
			}
			window_width_cols = winsize.ws_col;
			window_height_rows = winsize.ws_row;
			/* Just like the prom does */
			/* helios:/usr/src/3.5/sun/mon3/sys/finit.c */
			if (window_width_cols < 10)
				window_width_cols = 80;
			else if (window_width_cols > 120)
				window_width_cols = 120;
			if (window_height_rows < 10)
				window_height_rows = 34;
			else if (window_height_rows > 48)
				window_height_rows = 80;

		} else {
			/* Low res (and everything else) */
			window_width_cols = 80;
			window_height_rows = 34;
		}
	}

	/* If full screen, max out width and height */
	if (!default_window && MAX) {
		window_width_cols = width / font_width;
		window_height_rows = height / font_height;
	}

	/* get distance to top and bottom of font from font origin */
	font_ascent = -(font->pf_char)['a'].pc_home.y;
	font_descent = font_height - font_ascent;

	window_width_pix = window_width_cols * font_width;
	window_height_pix = window_height_rows * font_height;

	/* set window origin so it is perfectly centered on screen */
	origin_x = (width - window_width_pix) / 2;
	origin_y = (height - window_height_pix) / 2;
}

/**********************************************************************
The following graphics routines emulate the sun console termcap. See
console(4) for details
**********************************************************************/

/**********************************************************************
scroll screen up num_lines. The cursor position is shifted up with the
text. The screen may be shifted up past the top of the screen in which
case subsequent writes will not appear until the cursor is once again
in the window.
**********************************************************************/
scroll_up(num_lines)
	int num_lines;
{
	int ydiv;

	ytab -= num_lines;

	/* can't do better than erase the whole screen... */
	if (num_lines > window_height_rows)
		num_lines = window_height_rows;

	ydiv = font_height * num_lines;

	pr_rop(sp, origin_x, origin_y,
	    window_width_pix, window_height_pix - ydiv, PIX_SRC,
	    sp, origin_x, ydiv + origin_y);

	pr_rop(sp, origin_x, window_height_pix - ydiv + origin_y,
	    window_width_pix, ydiv, PIX_CLR,
	    sp, origin_x, window_height_pix - ydiv + origin_y);
}

/**********************************************************************
delete num chars starting from the current cursor position and moving
right. All chars up to the right border of the screen will be scrolled
left to fill the gap. White space appears on the right. Cursor
position is unchanged.
**********************************************************************/
delete_char(num)
	int num;
{
	int tox, fromx, y, width;

	/* can't delete more chars than there are */
	if (num > window_width_cols - xtab)
		num = window_width_cols - xtab;

	fromx = (xtab + num) * font_width;
	tox = xtab * font_width;
	y = ytab * font_height;
	width = num * font_width;

	pr_rop(sp, tox + origin_x, y + origin_y, window_width_pix - fromx,
	    font_height, PIX_SRC, sp, fromx + origin_x, y + origin_y);

	pr_rop(sp, window_width_pix - width + origin_x, y + origin_y, width,
	    font_height, PIX_CLR, sp,
	    window_width_pix - width + origin_x, y + origin_y);
}

/**********************************************************************
delete num lines starting from the current line and moving down. All
lines below will be scrolled up to fill the gap. White space appears
at the bottom of the window. Cursor position is unchanged.
**********************************************************************/
delete_line(num)
	int num;
{
	int fromy, toy, height;

	/* can't delete more lines than there are */
	if (num > window_height_rows - ytab)
		num = window_height_rows - ytab;

	fromy = (ytab + num) * font_height;
	toy = ytab * font_height;
	height = font_height * num;

	pr_rop(sp, origin_x, toy + origin_y, window_width_pix,
	    window_height_pix - fromy, PIX_SRC,
	    sp, origin_x, fromy + origin_y);

	pr_rop(sp, origin_x, window_height_pix - height + origin_y,
	    window_width_pix, height, PIX_CLR,
	    sp, origin_x, window_height_pix - height + origin_y);
}

/**********************************************************************
insert space for num chars starting at the current cursor position and
moving right. Chars which scroll off the right side of the screen are
lost. Cursor Position is unchanged.
**********************************************************************/
insert_char(num)
	int num;
{
	int tox, fromx, y;

	/* can't insert more chars than can fit */
	if (num > window_width_cols - xtab)
		num = window_width_cols - xtab;

	tox = (xtab + num) * font_width;
	fromx = xtab * font_width;
	y = ytab * font_height;

	pr_rop(sp, tox + origin_x, y + origin_y, window_width_pix - tox,
	    font_height, PIX_SRC, sp, fromx + origin_x, y + origin_y);

	pr_rop(sp, fromx + origin_x, y + origin_y, font_width * num,
	    font_height, PIX_CLR, sp, fromx + origin_x, y + origin_y);
}

/**********************************************************************
insert space for num lines starting at the current line and moving
down. Lines scrolled off the bottom of the screen are lost. The cursor
position is unchanged.
**********************************************************************/
insert_line(num)
	int num;
{
	int fromy, toy;

	/* can't insert more lines than can fit */
	if (num > window_height_rows - ytab)
		num = window_height_rows - ytab;

	toy = (ytab + num) * font_height;
	fromy = ytab * font_height;

	pr_rop(sp, origin_x, toy + origin_y, window_width_pix,
	    window_height_pix - toy, PIX_SRC,
	    sp, origin_x, fromy + origin_y);

	pr_rop(sp, origin_x, fromy + origin_y, window_width_pix,
	    font_height * num, PIX_CLR,
	    sp, origin_x, fromy + origin_y);
}

/**********************************************************************
clear to end of line starting at the current cursor position, which is
unchanged.
**********************************************************************/
cltoeol()
{
	int x;

	x = xtab * font_width;

	pr_rop(sp, x + origin_x, ytab * font_height + origin_y,
	    window_width_pix - x, font_height, PIX_CLR,
	    sp, x + origin_x, ytab * font_height + origin_y);
}

/**********************************************************************
clear to end of page starting at the current cursor position, which is
unchanged.
**********************************************************************/
cltoeop()
{
	int y;

	cltoeol();
	y = (ytab + 1) * font_height;

	pr_rop(sp, origin_x, y + origin_y,
	    window_width_pix, window_height_pix - y, PIX_CLR,
	    sp, origin_x, y + origin_y);
}

/**********************************************************************
This routine receives a string containing a sun console escape
sequence minus the initial escape, (e.g. [#1;#2H ), the length of the
string, a pointer to an array of ints in which to return the parsed
values, and the number of values to parse. The values in the string
are extracted and returned in the passed int array. If there are more
parameters in the string than requested in num_parms, the right-most
num_parms values will be returned. Parameters which correspond to
empty spaces in the escape sequence string will be unchanged on
return. See console(4) for complete details.
**********************************************************************/
parse_esc(buf, length, parms, num_parms)
	char *buf;
	int length, *parms, num_parms;
{
	int x10;

	length -= 2;
	num_parms--;

	do {
		x10 = 1;

		if (isdigit(buf[length]))
			parms[num_parms] = 0;

		while (isdigit(buf[length])) {
			parms[num_parms] += ((buf[length] - '0') * x10);
			x10 *= 10;
			length--;
		}
		while (buf[length] == ';') {
			num_parms--;
			length--;
		}
	}
	while (length > 0 && num_parms >= 0);
}

/**********************************************************************
Handle an insert-char escape sequence.

This routine is the same as many below it, so I will only document it
once. The values of the parameters are set to the appropriate default
value as per the spec (console(4) ), then the escape sequence string
is parsed to extract any parameters which are expected. Finally, the
parameters are passed to the appropriate handling routine.
**********************************************************************/
do_insert_char(buf, length)
	char *buf;
	int length;
{
	int parm[1], x, num_parms = 1;

	for (x = 0; x < num_parms; x++)
		parm[x] = 1;

	parse_esc(buf, length, parm, num_parms);

	insert_char(parm[0]);
}

/**********************************************************************
**********************************************************************/
do_delete_char(buf, length)
	char *buf;
	int length;
{
	int parm[1], x, num_parms = 1;

	for (x = 0; x < num_parms; x++)
		parm[x] = 1;

	parse_esc(buf, length, parm, num_parms);

	delete_char(parm[0]);
}

/**********************************************************************
**********************************************************************/
do_cursor_up(buf, length)
	char *buf;
	int length;
{
	int parm[1], x, num_parms = 1;

	for (x = 0; x < num_parms; x++)
		parm[x] = 1;

	parse_esc(buf, length, parm, num_parms);

	ytab -= parm[0];
	ytab = ytab < 0 ? 0 : ytab;
}

/**********************************************************************
**********************************************************************/
do_cursor_down(buf, length)
	char *buf;
	int length;
{
	int parm[1], x, num_parms = 1;

	for (x = 0; x < num_parms; x++)
		parm[x] = 1;

	parse_esc(buf, length, parm, num_parms);

	ytab += parm[0];
	ytab = ytab >= window_height_rows ? window_height_rows - 1 : ytab;
}

/**********************************************************************
**********************************************************************/
do_cursor_foreward(buf, length)
	char *buf;
	int length;
{
	int parm[1], x, num_parms = 1;

	for (x = 0; x < num_parms; x++)
		parm[x] = 1;

	parse_esc(buf, length, parm, num_parms);

	xtab += parm[0];
	if (xtab >= window_width_cols)
		xtab = window_width_cols - 1;
}

/**********************************************************************
**********************************************************************/
do_cursor_backward(buf, length)
	char *buf;
	int length;
{
	int parm[1], x, num_parms = 1;

	for (x = 0; x < num_parms; x++)
		parm[x] = 1;

	parse_esc(buf, length, parm, num_parms);

	xtab -= parm[0];
	if (xtab < 0)
		xtab = 0;
}

/**********************************************************************
**********************************************************************/
do_cursor_next_line(buf, length)
	char *buf;
	int length;
{
	int parm[1], x, num_parms = 1;

	for (x = 0; x < num_parms; x++)
		parm[x] = 1;

	parse_esc(buf, length, parm, num_parms);

	xtab = 0;
	ytab += parm[0];
	if (ytab >= window_height_rows)
		ytab = window_height_rows - 1;
}

/**********************************************************************
**********************************************************************/
do_hvtab(buf, length)
	char *buf;
	int length;
{
	int parm[2], x, num_parms = 2;

	for (x = 0; x < num_parms; x++)
		parm[x] = 1;

	parse_esc(buf, length, parm, num_parms);

	xtab = parm[1] - 1;
	ytab = parm[0] - 1;
}

/**********************************************************************
**********************************************************************/
do_insert_line(buf, length)
	char *buf;
	int length;
{
	int parm[1], x, num_parms = 1;

	for (x = 0; x < num_parms; x++)
		parm[x] = 1;

	parse_esc(buf, length, parm, num_parms);

	insert_line(parm[0]);
}

/**********************************************************************
**********************************************************************/
do_delete_line(buf, length)
	char *buf;
	int length;
{
	int parm[1], x, num_parms = 1;

	for (x = 0; x < num_parms; x++)
		parm[x] = 1;

	parse_esc(buf, length, parm, num_parms);

	delete_line(parm[0]);
}

/**********************************************************************
**********************************************************************/
do_set_graphic(buf, length)
	char *buf;
	int length;
{
	int parm[1], x, num_parms = 1;

	for (x = 0; x < num_parms; x++)
		parm[x] = 0;

	parse_esc(buf, length, parm, num_parms);

	STANDOUT_MODE = parm[0];
}

/**********************************************************************
**********************************************************************/
do_set_scroll_mode(buf, length)
	char *buf;
	int length;
{
	int parm[1], x, num_parms = 1;

	for (x = 0; x < num_parms; x++)
		parm[x] = 0;

	parse_esc(buf, length, parm, num_parms);

	/* who needs it... */
}

/**********************************************************************
handle the bell.
**********************************************************************/
do_bell()
{
	int x, i;

	if (VISUAL_BELL) {

		/* invert the screen twice */

		for (x = 0; x < 2; x++)
			pr_rop(sp, origin_x, origin_y, window_width_pix, window_height_pix,
			    PIX_NOT(PIX_DST), sp, origin_x, origin_y);
	} else {

		/* fd 0 refers to /dev/kbd */

		i = KBD_CMD_BELL;
		if (ioctl(0, KIOCCMD, &i) < 0) {
			perror("tconsole: KBD_CMD_BELL");
			exit(1);
		}
		usleep(100000);

		i = KBD_CMD_NOBELL;
		if (ioctl(0, KIOCCMD, &i) < 0) {
			perror("tconsole: KBD_CMD_NOBELL");
			exit(1);
		}
	}
}

/**********************************************************************
Handle a sun console escape sequence. This routine is called after an
initial ESC has been read. First the rest of the escape sequence is
read into a buffer. get_char() is called and instructed to read() more
data in blocking-io mode if the input buffer is empty, so we don't
return until the entire escape sequence has been processed.  The final
char in the sequence indicates the type of esc request, and the
appropriate handler is called.
**********************************************************************/
handle_esc(pty)
	int pty;
{
	int c, offset;
	char buf[128];

	do {
		offset = 0;
		do {

			/* read a char with blocking io */

			c = get_char(pty, READ | RET_DATA);
			buf[offset++] = c;
		}
		while (isdigit(c) || c == ';' || c == '[');
	}

	/* start over if a second ESC was encountered, i.e. throw out the first
	 * sequence (must have been garbled) */

	while (c == ESC);


	buf[offset] = 0;	/* make it a string with null termination */

	switch (c) {
	case '@':
		do_insert_char(buf, offset);
		break;
	case 'A':
		do_cursor_up(buf, offset);
		break;
	case 'B':
		do_cursor_down(buf, offset);
		break;
	case 'C':
		do_cursor_foreward(buf, offset);
		break;
	case 'D':
		do_cursor_backward(buf, offset);
		break;
	case 'E':
		do_cursor_next_line(buf, offset);
		break;
	case 'f':
	case 'H':
		do_hvtab(buf, offset);
		break;
	case 'J':
		cltoeop();
		break;
	case 'K':
		cltoeol();
		break;
	case 'L':
		do_insert_line(buf, offset);
		break;
	case 'M':
		do_delete_line(buf, offset);
		break;
	case 'P':
		do_delete_char(buf, offset);
		break;
	case 'm':
		do_set_graphic(buf, offset);
		break;
	case 'p':
		STANDOUT_MODE = 0;
		break;
	case 'q':
		STANDOUT_MODE = 1;
		break;
	case 'r':
		do_set_scroll_mode(buf, offset);
		break;
	case 's':
		STANDOUT_MODE = 0;
		SCROLL_MODE = 0;
		break;
	}
}

/**********************************************************************
Handle jump scrolling
**********************************************************************/
jump_scroll(pty)
	int pty;
{
	int num_NL;

	/* get the number of NL chars in the output buffer. If the output
	 * buffer is empty, there are none. */

	if ((num_NL = get_char(pty, RET_NL)) < 0)
		num_NL = 0;

	/* plus one for the NL that sent us into this routine in the first
	 * place */

	num_NL++;

	/* scroll the window as needed */

	scroll_up(num_NL);
}

/**********************************************************************
Handle a non-printable character other than ESC.
**********************************************************************/
handle_ctrl_char(c, pty)
	char c;
	int pty;
{
	switch (c) {

	case CR:
		xtab = 0;
		break;

	case BS:
		if (xtab > 0)
			xtab--;
		break;

	case VT:
		if (ytab > 0)
			ytab--;
		break;

	case NP:
		ytab = xtab = 0;
		cltoeop();
		break;

	case NL:
		ytab++;
		if (ytab >= window_height_rows)
			jump_scroll(pty);
		break;
	case BEL:
		do_bell();
		break;

	case HT:
		xtab = (xtab + 8) & ~7;
		if (xtab >= window_width_cols)
			xtab = window_width_cols - 1;
		break;
	}
}

/**********************************************************************
Print a string to the window starting at the current cursor position
and updating the cursor position to the end of the string. It is
implicit that the caller will not send a string long enough to run off
the right side of the window. If the vertical tab is not within the
window, the cursor position will be updated, but no printing will
occur, since very fast jump scrolling may cause the cursor to be
scrolled up past the top of the window.
**********************************************************************/
print_string(buf, length)
	char *buf;
	int length;
{
	register int rop, x, y;

	if (ytab >= 0) {
		x = xtab * font_width + origin_x;
		y = ytab * font_height + font_ascent + origin_y + 1;
		pr_text(sp, x, y, PIX_SRC, font, buf);
		if (STANDOUT_MODE) {
			x = xtab * font_width + origin_x;
			y = ytab * font_height + origin_y;
			pr_rop(sp, x, y, font_width * length, font_height,
			    PIX_NOT(PIX_DST), sp, x, y);
		}
	}
	xtab += length;
	if (xtab >= window_width_cols) {
		xtab = 0;
		ytab++;
	}
	if (ytab >= window_height_rows)
		scroll_up(1);
}

/**********************************************************************
Draw a cursor by inverting the character at the current cursor
position. This is done by xor'ing an inverted space with the char at
the current location.
**********************************************************************/
putcursor()
{
	register int x, y;

	x = xtab * font_width + origin_x;
	y = ytab * font_height + origin_y;
	pr_rop(sp, x, y, font_width, font_height, PIX_NOT(PIX_DST), sp, x, y);
}

/*****************************************************************************
clear the screen and home the cursor.
*****************************************************************************/
init()
{

	pr_rop(sp, 0, 0, sp->pr_size.x, sp->pr_size.y, PIX_CLR,
	    sp, 0, 0);

	xtab = ytab = 0;
}
